from typing import Dict, Collection

from faker import Faker

from ..config import ColumnOption
from .column import Column
from .tools.tool import ToolContext


class Row:
    def __init__(self, table: str, cols: Dict[str, Collection[ColumnOption]]):
        self.table = table
        self.cols = [Column(name, cols[name]) for name in cols]

    def exec(self, index: int, lang: str, fake: Faker, rel: Dict[str, Dict] = None):
        if rel is None:
            rel = {}

        ctx: ToolContext = {
            'table': self.table,
            'col': '',
            'index': index,
            'lang': lang,
            'fake': fake,
            'curr': {},
            'rel': rel
        }

        row = {}
        for col in self.cols:
            ctx['col'] = col.name
            row[col.name] = col.exec(ctx)
            ctx['curr'] = dict(ctx['curr'], **row)
        return row



from typing import Any

from .tool import ITool, ToolContext


class ParamTool(ITool):
    def exec(self, ctx: ToolContext, args: Any = None) -> Any:
        assert ('key' in args), "No key specified for args" + self.to_string(ctx)
        if isinstance(args['key'], list):
            curr = args
            for key in args['key']:
                assert (key in curr), "Specified key not found for args" + self.to_string(ctx)
                curr = curr[key]
            return curr

        assert (args['key'] in args), "Specified key not found for args" + self.to_string(ctx)
        return args[args['key']]



import json
from typing import Any, TypedDict, Dict

from faker import Faker


class ToolContext(TypedDict):
    table: str
    col: str
    index: int
    lang: str
    fake: Faker
    curr: Dict
    rel: Dict[str, Dict]


class ITool:
    def exec(self, ctx: ToolContext, args: Any = None) -> Any:
        """Overrides InformalParserInterface.load_data_source()"""
        pass

    @staticmethod
    def to_string(ctx: ToolContext):
        return f" {ctx['table']} {ctx['col']} {ctx['index']} {ctx['lang']} {json.dumps(ctx['curr'])} {json.dumps(ctx['rel'])}"



from typing import Any

from .tool import ITool, ToolContext


class ContextTool(ITool):
    def exec(self, ctx: ToolContext, args: Any = None) -> Any:
        assert ('key' in args), "No key specified for context" + self.to_string(ctx)
        if isinstance(args['key'], list):
            curr = ctx
            for key in args['key']:
                assert (key in curr), "Specified key not found for context" + self.to_string(ctx)
                curr = curr[key]
            return curr

        assert (args['key'] in ctx), "Specified key not found for context" + self.to_string(ctx)
        return ctx[args['key']]



from .faker import FakerTool
from .random import RandomTool
from .context import ContextTool
from .constant import ConstantTool
from .param import ParamTool
from .string import JoinTool, SlugifyTool

tools = {
    'constant': ConstantTool(),
    'param': ParamTool(),
    'context': ContextTool(),
    'faker': FakerTool(),
    'random': RandomTool(),
    'string.join': JoinTool(),
    'string.slugify': SlugifyTool(),
}



from typing import Any

from ..tool import ITool, ToolContext


class JoinTool(ITool):
    def exec(self, ctx: ToolContext, args: Any = None) -> Any:
        assert ('sep' in args), "No separator specified for join function" + self.to_string(ctx)
        assert ('parts' in args), "No parts specified for join function" + self.to_string(ctx)
        return args['sep'].join(args['parts'])



import random
from csv import DictWriter
from typing import Dict, Callable, Collection

from faker import Faker

from ..config import TableConfig
from .row import Row


class Table:
    def __init__(self, config: TableConfig, fakes: Dict[str, Faker], default_lang: str = 'en'):
        self.option = config['option']
        self.row = Row(config['name'], config['cols'])
        self.default = config['default'] if 'default' in config else []
        self.index = len(self.default)
        self.default_lang = default_lang
        self.fakes = fakes

    def exec(self, writer: DictWriter, get_rel: Callable[[str], Collection[Dict]]):
        # Write default rows
        for row in self.default:
            writer.writerow(row)

        if self.option['type'] == 'simple':
            self.simple(writer)

        if self.option['type'] == 'multiply':
            self.multiply(writer, get_rel)

        if self.option['type'] == 'relation':
            self.relation(writer, get_rel)

    def simple(self, writer: DictWriter):
        for i in range(1, self.option['args']['count'] + 1):
            self.index += 1
            writer.writerow(self.row.exec(self.index, self.default_lang, self.fakes[self.default_lang], None))

    def multiply(self, writer: DictWriter, get_rel: Callable[[str], Collection[Dict]]):
        rel_rows = get_rel(self.option['args']['table'])
        if 'where' in self.option['args']:
            rel_rows = self.get_valid_rows(rel_rows, self.option['args']['where'])
        for row in rel_rows:
            if 'var' in self.option['args'] and random.random() < self.option['args']['var']:
                continue

            rel = {self.option['args']['table']: row}
            for i in range(self.option['args']['count']):
                self.index += 1

                selected_lang = self.default_lang
                if 'languages' in self.option['args']:
                    selected_lang = self.option['args']['languages'][i]
                writer.writerow(self.row.exec(self.index, selected_lang, self.fakes[selected_lang], rel))

    def relation(self, writer: DictWriter, get_rel: Callable[[str], Collection[Dict]]):
        left_rel_rows = get_rel(self.option['args']['left']['table'])
        if 'where' in self.option['args']['left']:
            left_rel_rows = self.get_valid_rows(left_rel_rows, self.option['args']['left']['where'])

        right_rel_rows = get_rel(self.option['args']['right']['table'])
        if 'where' in self.option['args']['right']:
            right_rel_rows = self.get_valid_rows(right_rel_rows, self.option['args']['right']['where'])

        if self.option['args']['count'] == 'cover':
            for row in left_rel_rows:
                right_row = random.choice(right_rel_rows)
                self.index += 1

                rel = {
                    self.option['args']['left']['table']: row,
                    self.option['args']['right']['table']: right_row
                }
                writer.writerow(self.row.exec(self.index, self.default_lang, self.fakes[self.default_lang], rel))
        else:
            for i in range(self.option['args']['count']):
                left_row = random.choice(left_rel_rows)
                right_row = random.choice(right_rel_rows)
                self.index += 1

                rel = {
                    self.option['args']['left']['table']: left_row,
                    self.option['args']['right']['table']: right_row
                }
                writer.writerow(self.row.exec(self.index, self.default_lang, self.fakes[self.default_lang], rel))

    @staticmethod
    def get_valid_rows(rel_rows: Collection, conds: Collection):
        valid_rows = []
        for target in rel_rows:
            eligible = True
            for cond in conds:
                for key in cond:
                    if target[key] != cond[key]:
                        eligible = False
            if eligible:
                valid_rows.append(target)
        return valid_rows



import random
from typing import Any

from .tool import ITool, ToolContext


class RandomTool(ITool):
    def exec(self, ctx: ToolContext, args: Any = None) -> Any:
        assert hasattr(random, args['func']), "Specified function must be in random" + self.to_string(ctx)
        func = getattr(random, args['func'])
        params = args['params'] if 'params' in args else {}
        return func(**params)



from src import Lokii
from src.config import LokiConfig
import unittest


class FakerToolTest(unittest.TestCase):

    def test_pass(self):
        config: LokiConfig = {
            'languages': ['en'],
            'tables': [
                {
                    'name': 'test',
                    'option': {
                        'type': 'simple',
                        'args': {
                            'count': 5
                        }
                    },
                    'cols': {
                        'rand_email': [
                            {
                                'type': 'random',
                                'args': {
                                    'func': 'random'
                                }
                            }
                        ],
                        'rand_paragraph': [
                            {
                                'type': 'random',
                                'args': {
                                    'func': 'choice',
                                    'params': {
                                        'seq': [1, 2, 3]
                                    }
                                }
                            }
                        ]
                    }
                }
            ]
        }
        Lokii(config).generate()



from typing import Any
from slugify import slugify

from ..tool import ITool, ToolContext


class SlugifyTool(ITool):
    def exec(self, ctx: ToolContext, args: Any = None) -> Any:
        return slugify(args)



from typing import Collection

from ..config import ColumnOption
from .tools import tools
from .tools.tool import ToolContext


class Column:
    def __init__(self, name: str, data: Collection[ColumnOption]):
        self.name = name
        self.options = data

    def exec(self, ctx: ToolContext):
        calc = []
        for option in self.options:
            args = option['args'] if 'args' in option else {}

            assert ('type' in option), "Type not specified for column option! " + self.name
            assert (option['type'] in tools), "Tool can not be found! " + option['type']

            if 'calc' in option:
                if isinstance(option['calc']['keys'], list):
                    calc_values = []
                    for index in option['calc']['keys']:
                        calc_values.append(calc[index])
                else:
                    calc_values = calc[option['calc']['keys']]

                if 'as' in option['calc']:
                    args[option['calc']['as']] = calc_values
                else:
                    args = calc_values
            calc.append(tools[option['type']].exec(ctx, args))

        return calc[-1:][0] if len(calc) > 0 else None
